@@ -163,13 +163,13 @@ codec_choose_forced (struct codec *codec, int flags)
 static bool
 codec_choose_arm (struct codec *codec)
 {
-#if (defined(__ARM_NEON__) || defined(__ARM_NEON)) && ((defined(__aarch64__) && HAVE_NEON64) || HAVE_NEON32)
+#if HAVE_NEON64 || ((defined(__ARM_NEON__) || defined(__ARM_NEON)) && HAVE_NEON32)
 
 	// Unfortunately there is no portable way to check for NEON
 	// support at runtime from userland in the same way that x86
 	// has cpuid, so just stick to the compile-time configuration:
 
-	#if defined(__aarch64__) && HAVE_NEON64
+	#if HAVE_NEON64
 	codec->enc = base64_stream_encode_neon64;
 	codec->dec = base64_stream_decode_neon64;
 	#else
